from typing import List

from aidial_client._internal_types._http_request import FinalRequestOptions
from aidial_client.resources.base import AsyncResource, Resource
from aidial_client.types.deployment import Deployment, DeploymentsResponse


class Deployments(Resource):
    def _list_raw(self) -> DeploymentsResponse:
        return self.http_client.request(
            cast_to=DeploymentsResponse,
            options=FinalRequestOptions(method="GET", url="openai/deployments"),
        )

    def list(self) -> List[Deployment]:
        return self._list_raw().data


class AsyncDeployments(AsyncResource):
    async def _list_raw(self) -> DeploymentsResponse:
        return await self.http_client.request(
            cast_to=DeploymentsResponse,
            options=FinalRequestOptions(method="GET", url="openai/deployments"),
        )

    async def list(self) -> List[Deployment]:
        return (await self._list_raw()).data



from pathlib import PurePosixPath
from typing import Literal, Optional, Union, cast, get_args
from urllib.parse import urljoin, urlparse

from aidial_client._compatibility.pydantic_v1 import BaseModel
from aidial_client._constants import API_PREFIX
from aidial_client._exception import InvalidDialURLError, NotDialURLError
from aidial_client.helpers._url import enforce_trailing_slash

StorageResourceType = Literal["files", "conversations", "prompts"]


def _is_directory(s: str) -> bool:
    return s[-1] == "/"


class DialStorageResource(BaseModel):
    resource_type: StorageResourceType

    """Bucket name, like 'my-bucket'"""
    bucket: str

    """Absolute url, like 'https://dial.core/v1/files/my-bucket/my-file.txt'"""
    absolute_url: str

    """Relative url, like '/v1/files/my-bucket/my-file.txt'"""
    relative_url: str

    """Path without api prefix, like 'files/my-bucket/my-folder/my-file.txt'"""
    api_path: str

    """Path without bucket, like my-folder/'my-file.txt'"""
    bucket_path: str

    """
    Filename, like 'my-file.txt'
    None for a directory
    """
    filename: Optional[str] = None


def safe_parse_storage_resource(
    *,
    url: str,
    dial_api_url: str,
    expected_resource_type: Optional[StorageResourceType] = None,
) -> Union[DialStorageResource, NotDialURLError, InvalidDialURLError]:
    """
    Parse the storage resource from the URL, that could be
    1. Absolute: "https://dial.core/v1/files/my-bucket/my-file.txt"
    2. Relative to API prefix: "files/my-bucket/my-file.txt"
    """
    dial_api_url = enforce_trailing_slash(dial_api_url)
    if url.startswith("/"):
        return InvalidDialURLError(f"Root-relative URL is forbidden: {url}")
    if url.startswith(API_PREFIX):
        return InvalidDialURLError(
            f"API prefix as relative part is not allowed: {url}"
        )

    absolute_url = urljoin(dial_api_url, url)
    url_parsed = urlparse(absolute_url)
    dial_api_parsed = urlparse(dial_api_url)
    if url_parsed.netloc != dial_api_parsed.netloc:
        return NotDialURLError(message=f"Provided URL is not DIAL URL: {url}")
    try:
        url_path = PurePosixPath(url_parsed.path)
        api_path = url_path.relative_to(dial_api_parsed.path)
    except ValueError:
        return InvalidDialURLError(
            f"Provided URL path {url_parsed.path} does not match with"
            f" DIAL API URL {dial_api_parsed.path}"
        )

    resource_path = api_path.parents[len(api_path.parents) - 2]
    parsed_resource_type = str(resource_path)

    if parsed_resource_type not in get_args(StorageResourceType):
        return InvalidDialURLError(
            f"Invalid resource type: {parsed_resource_type}"
        )
    # If user provided expected resource type, check it
    if (
        expected_resource_type is not None
        and parsed_resource_type != expected_resource_type
    ):
        return InvalidDialURLError(
            f"Invalid resource type for URL: {url}\n"
            f"Expected: {expected_resource_type}, got: {parsed_resource_type}"
        )

    if len(api_path.parents) < 3:
        return InvalidDialURLError(f"Missing bucket in URL: {url}")

    bucket_path = api_path.parents[len(api_path.parents) - 3]
    return DialStorageResource(
        resource_type=cast(StorageResourceType, parsed_resource_type),
        absolute_url=absolute_url,
        api_path=str(api_path),
        bucket=str(bucket_path.relative_to(resource_path)),
        bucket_path=str(api_path.relative_to(bucket_path)),
        relative_url=str(url_path),
        filename=url_path.name if not _is_directory(url) else None,
    )


def parse_storage_resource(
    *,
    url: str,
    dial_api_url: str,
    expected_resource_type: Optional[StorageResourceType] = None,
) -> DialStorageResource:
    result = safe_parse_storage_resource(
        url=url,
        dial_api_url=dial_api_url,
        expected_resource_type=expected_resource_type,
    )
    if isinstance(result, (NotDialURLError, InvalidDialURLError)):
        raise result
    return result


class DialStorageResourceMixin(BaseModel):
    """
    Mixin class for resources that are using DIAL storage:
    - /v1/files
    - /v1/conversations
    - /v1/prompts
    """

    resource_type: StorageResourceType
    dial_api_url: str

    def get_storage_resource(self, url: str) -> DialStorageResource:
        """
        Get the storage resource object from the URL
        Args:
            url (str): The URL to be processed.
        Returns:
            DialStorageResource: The storage resource object
        """
        return parse_storage_resource(
            url=url,
            dial_api_url=self.dial_api_url,
            expected_resource_type=self.resource_type,
        )

    def get_api_path(self, url: str) -> str:
        """
        Convert URL, that could relative or absolute, to relative URL
        """
        return self.get_storage_resource(url).api_path

    def get_display_name(self, url: str) -> str:
        """
        Get the display name of the resource from the URL
        """
        return self.get_storage_resource(url).bucket_path



from aidial_client.resources.deployments import AsyncDeployments, Deployments
from aidial_client.resources.metadata import AsyncMetadata, Metadata

from .application import Application, AsyncApplication
from .bucket import AsyncBucket, Bucket
from .chat import AsyncChat, Chat
from .files import AsyncFiles, Files

__all__ = [
    "Chat",
    "AsyncChat",
    "Bucket",
    "AsyncBucket",
    "Files",
    "AsyncFiles",
    "AsyncDeployments",
    "Deployments",
    "AsyncMetadata",
    "Metadata",
    "Application",
    "AsyncApplication",
]



from io import BufferedReader
from typing import (
    IO,
    Any,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
    final,
)

from httpx import Timeout

from aidial_client._compatibility.pydantic_v1 import BaseModel
from aidial_client._internal_types._defaults import NOT_GIVEN, NotGiven

FileContent = Union[
    IO[bytes],
    bytes,
    str,
    # Somehow, pydantic doesn't recognize result of open('...', 'rb') as IO[bytes]
    # even though BufferedReader is a subclass of IO[bytes]
    BufferedReader,
]
FileTypes = Union[
    # file (or bytes)
    FileContent,
    # (filename, file (or bytes))
    Tuple[Optional[str], FileContent],
    # (filename, file (or bytes), content_type)
    Tuple[Optional[str], FileContent, Optional[str]],
    # (filename, file (or bytes), content_type, headers)
    Tuple[Optional[str], FileContent, Optional[str], Mapping[str, str]],
]

Params = Mapping[str, Any]
Headers = Mapping[str, Any]
Data = Mapping[str, Any]
RequestFiles = Union[Mapping[str, FileTypes], Sequence[Tuple[str, FileTypes]]]


@final
class FinalRequestOptions(BaseModel):
    class Config:
        arbitrary_types_allowed: bool = True

    method: Literal["GET", "PUT", "POST", "DELETE"]
    url: str
    params: Optional[Params] = None
    headers: Optional[Headers] = None
    max_retries: Union[int, NotGiven] = NOT_GIVEN
    timeout: Union[float, Timeout, NotGiven, None] = NOT_GIVEN
    files: Optional[RequestFiles] = None
    json_data: Optional[Data] = None

    def get_max_retries(self, max_retries: int) -> int:
        if isinstance(self.max_retries, NotGiven):
            return max_retries
        return self.max_retries

    def get_timeout(
        self, timeout: Union[float, Timeout, None]
    ) -> Union[float, Timeout, None]:
        if isinstance(self.timeout, NotGiven):
            return timeout
        return self.timeout



from typing import (
    Any,
    AsyncIterable,
    Dict,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Union,
    cast,
    overload,
)

import openai
from openai import AsyncStream as OpenaiAsyncStream
from openai import Stream as OpenaiStream
from openai.types.chat import ChatCompletion as OpenaiChatCompletion
from openai.types.chat import ChatCompletionChunk as OpenaiChatCompletionChunk
from pydantic import StrictStr

from aidial_client._compatibility.openai import Omit
from aidial_client._utils._dict import remove_none
from aidial_client._utils._openai import (
    convert_openai_async_stream,
    convert_openai_error,
    convert_openai_response,
    convert_openai_stream,
)
from aidial_client.resources.base import AsyncResource, Resource
from aidial_client.types.chat import (
    Addon,
    ChatCompletionChunk,
    ChatCompletionResponse,
    FunctionCallSpecParam,
    FunctionParam,
    Message,
    ToolCallSpecParam,
    ToolParam,
)
from aidial_client.types.chat.request import ChatCompletionRequestCustomFields


class ChatCompletions(Resource):
    default_api_version: Optional[str] = None
    openai_client: openai.AzureOpenAI

    @overload
    def create(
        self,
        *,
        deployment_name: str,
        messages: List[Message],
        stream: Literal[True],
        api_version: Optional[str] = None,
        model: Optional[str] = None,
        functions: Union[List[FunctionParam], None] = None,
        function_call: Union[
            Union[Literal["none", "auto"], FunctionCallSpecParam], None
        ] = None,
        tools: Union[List[ToolParam], None] = None,
        tool_choice: Union[
            Union[Literal["none", "auto"], ToolCallSpecParam], None
        ] = None,
        addons: Union[Addon, None] = None,
        temperature: Union[float, None] = None,
        top_p: Union[float, None] = None,
        n: Union[int, None] = None,
        stop: Union[Union[str, List[str]], None] = None,
        max_tokens: Union[int, None] = None,
        max_prompt_tokens: Union[Union[Literal["infinity"], int], None] = None,
        presence_penalty: Union[float, None] = None,
        frequency_penalty: Union[float, None] = None,
        logit_bias: Union[Dict, None] = None,
        seed: Union[int, None] = None,
        user: Union[str, None] = None,
        custom_fields: Union[ChatCompletionRequestCustomFields, None] = None,
        logprobs: Union[bool, None] = None,
        top_logprobs: Union[int, None] = None,
        # Extra params
        extra_body: Optional[Dict[str, Any]] = None,
        extra_headers: Optional[Mapping[StrictStr, StrictStr]] = None,
        extra_params: Optional[Dict[str, Any]] = None,
    ) -> Iterable[ChatCompletionChunk]: ...

    @overload
    def create(
        self,
        *,
        deployment_name: str,
        messages: List[Message],
        stream: Literal[False],
        api_version: Optional[str] = None,
        model: Optional[str] = None,
        functions: Union[List[FunctionParam], None] = None,
        function_call: Union[
            Union[Literal["none", "auto"], FunctionCallSpecParam], None
        ] = None,
        tools: Union[List[ToolParam], None] = None,
        tool_choice: Union[
            Union[Literal["none", "auto"], ToolCallSpecParam], None
        ] = None,
        addons: Union[Addon, None] = None,
        temperature: Union[float, None] = None,
        top_p: Union[float, None] = None,
        n: Union[int, None] = None,
        stop: Union[Union[str, List[str]], None] = None,
        max_tokens: Union[int, None] = None,
        max_prompt_tokens: Union[Union[Literal["infinity"], int], None] = None,
        presence_penalty: Union[float, None] = None,
        frequency_penalty: Union[float, None] = None,
        logit_bias: Union[Dict, None] = None,
        seed: Union[int, None] = None,
        user: Union[str, None] = None,
        custom_fields: Union[ChatCompletionRequestCustomFields, None] = None,
        logprobs: Union[bool, None] = None,
        top_logprobs: Union[int, None] = None,
        # Extra params
        extra_body: Optional[Dict[str, Any]] = None,
        extra_headers: Optional[Mapping[StrictStr, StrictStr]] = None,
        extra_params: Optional[Dict[str, Any]] = None,
    ) -> ChatCompletionResponse: ...

    def create(
        self,
        *,
        deployment_name: str,
        messages: List[Message],
        api_version: Optional[str] = None,
        stream: bool = False,
        model: Optional[str] = None,
        functions: Union[List[FunctionParam], None] = None,
        function_call: Union[
            Union[Literal["none", "auto"], FunctionCallSpecParam], None
        ] = None,
        tools: Union[List[ToolParam], None] = None,
        tool_choice: Union[
            Union[Literal["none", "auto"], ToolCallSpecParam], None
        ] = None,
        addons: Union[Addon, None] = None,
        temperature: Union[float, None] = None,
        top_p: Union[float, None] = None,
        n: Union[int, None] = None,
        stop: Union[Union[str, List[str]], None] = None,
        max_tokens: Union[int, None] = None,
        max_prompt_tokens: Union[Union[Literal["infinity"], int], None] = None,
        presence_penalty: Union[float, None] = None,
        frequency_penalty: Union[float, None] = None,
        logit_bias: Union[Dict, None] = None,
        seed: Union[int, None] = None,
        user: Union[str, None] = None,
        custom_fields: Union[ChatCompletionRequestCustomFields, None] = None,
        logprobs: Union[bool, None] = None,
        top_logprobs: Union[int, None] = None,
        # Extra params
        extra_body: Optional[Dict[str, Any]] = None,
        extra_headers: Optional[Mapping[StrictStr, StrictStr]] = None,
        extra_params: Optional[Dict[str, Any]] = None,
    ) -> Union[ChatCompletionResponse, Iterable[ChatCompletionChunk]]:

        model = model or deployment_name
        extra_body = extra_body or {}
        extra_headers = extra_headers or {}
        extra_params = extra_params or {}

        input_params = remove_none(
            {
                "messages": messages,
                "model": model,
                "frequency_penalty": frequency_penalty,
                "function_call": function_call,
                "functions": functions,
                "logit_bias": logit_bias,
                "max_tokens": max_tokens,
                "n": n,
                "presence_penalty": presence_penalty,
                "seed": seed,
                "stop": stop,
                "stream": stream,
                "temperature": temperature,
                "tool_choice": tool_choice,
                "tools": tools,
                "top_p": top_p,
                "user": user,
                "addons": addons,
                "max_prompt_tokens": max_prompt_tokens,
                "custom_fields": custom_fields,
                "logprobs": logprobs,
                "top_logprobs": top_logprobs,
                "extra_body": extra_body,
                "extra_query": {
                    "api-version": (
                        api_version or self.default_api_version or Omit()
                    )
                },
                "extra_headers": {
                    # We use Omit to override openai client auth headers
                    **{"Authorization": Omit(), "api-key": Omit()},
                    **(self.http_client.auth_headers()),
                    **extra_headers,
                },
            }
        )
        try:
            openai_response = self.openai_client.chat.completions.create(
                **input_params,
            )
            openai_response = cast(
                Union[
                    OpenaiChatCompletion,
                    OpenaiStream[OpenaiChatCompletionChunk],
                ],
                openai_response,
            )
        except openai.APIError as err:
            raise convert_openai_error(err)

        if isinstance(openai_response, OpenaiChatCompletion):
            return convert_openai_response(openai_response)
        else:
            return convert_openai_stream(openai_response)


class AsyncChatCompletions(AsyncResource):
    default_api_version: Optional[str] = None
    openai_client: openai.AsyncAzureOpenAI

    @overload
    async def create(
        self,
        *,
        deployment_name: str,
        messages: List[Message],
        stream: Literal[True],
        api_version: Optional[str] = None,
        model: Optional[str] = None,
        functions: Union[List[FunctionParam], None] = None,
        function_call: Union[
            Union[Literal["none", "auto"], FunctionCallSpecParam], None
        ] = None,
        tools: Union[List[ToolParam], None] = None,
        tool_choice: Union[
            Union[Literal["none", "auto"], ToolCallSpecParam], None
        ] = None,
        addons: Union[Addon, None] = None,
        temperature: Union[float, None] = None,
        top_p: Union[float, None] = None,
        n: Union[int, None] = None,
        stop: Union[Union[str, List[str]], None] = None,
        max_tokens: Union[int, None] = None,
        max_prompt_tokens: Union[Union[Literal["infinity"], int], None] = None,
        presence_penalty: Union[float, None] = None,
        frequency_penalty: Union[float, None] = None,
        logit_bias: Union[Dict, None] = None,
        seed: Union[int, None] = None,
        user: Union[str, None] = None,
        custom_fields: Union[ChatCompletionRequestCustomFields, None] = None,
        # Extra params
        extra_body: Optional[Dict[str, Any]] = None,
        extra_headers: Optional[Mapping[StrictStr, StrictStr]] = None,
        extra_params: Optional[Dict[str, Any]] = None,
    ) -> AsyncIterable[ChatCompletionChunk]: ...

    @overload
    async def create(
        self,
        *,
        deployment_name: str,
        messages: List[Message],
        stream: Literal[False],
        api_version: Optional[str] = None,
        model: Optional[str] = None,
        functions: Union[List[FunctionParam], None] = None,
        function_call: Union[
            Union[Literal["none", "auto"], FunctionCallSpecParam], None
        ] = None,
        tools: Union[List[ToolParam], None] = None,
        tool_choice: Union[
            Union[Literal["none", "auto"], ToolCallSpecParam], None
        ] = None,
        addons: Union[Addon, None] = None,
        temperature: Union[float, None] = None,
        top_p: Union[float, None] = None,
        n: Union[int, None] = None,
        stop: Union[Union[str, List[str]], None] = None,
        max_tokens: Union[int, None] = None,
        max_prompt_tokens: Union[Union[Literal["infinity"], int], None] = None,
        presence_penalty: Union[float, None] = None,
        frequency_penalty: Union[float, None] = None,
        logit_bias: Union[Dict, None] = None,
        seed: Union[int, None] = None,
        user: Union[str, None] = None,
        custom_fields: Union[ChatCompletionRequestCustomFields, None] = None,
        logprobs: Union[bool, None] = None,
        top_logprobs: Union[int, None] = None,
        # Extra params
        extra_body: Optional[Dict[str, Any]] = None,
        extra_headers: Optional[Mapping[StrictStr, StrictStr]] = None,
        extra_params: Optional[Dict[str, Any]] = None,
    ) -> ChatCompletionResponse: ...

    async def create(
        self,
        *,
        deployment_name: str,
        messages: List[Message],
        api_version: Optional[str] = None,
        stream: bool = False,
        model: Optional[str] = None,
        functions: Union[List[FunctionParam], None] = None,
        function_call: Union[
            Union[Literal["none", "auto"], FunctionCallSpecParam], None
        ] = None,
        tools: Union[List[ToolParam], None] = None,
        tool_choice: Union[
            Union[Literal["none", "auto"], ToolCallSpecParam], None
        ] = None,
        addons: Union[Addon, None] = None,
        temperature: Union[float, None] = None,
        top_p: Union[float, None] = None,
        n: Union[int, None] = None,
        stop: Union[Union[str, List[str]], None] = None,
        max_tokens: Union[int, None] = None,
        max_prompt_tokens: Union[Union[Literal["infinity"], int], None] = None,
        presence_penalty: Union[float, None] = None,
        frequency_penalty: Union[float, None] = None,
        logit_bias: Union[Dict, None] = None,
        seed: Union[int, None] = None,
        user: Union[str, None] = None,
        custom_fields: Union[ChatCompletionRequestCustomFields, None] = None,
        logprobs: Union[bool, None] = None,
        top_logprobs: Union[int, None] = None,
        # Extra params
        extra_body: Optional[Dict[str, Any]] = None,
        extra_headers: Optional[Mapping[StrictStr, StrictStr]] = None,
        extra_params: Optional[Dict[str, Any]] = None,
    ) -> Union[ChatCompletionResponse, AsyncIterable[ChatCompletionChunk]]:
        model = model or deployment_name
        extra_body = extra_body or {}
        extra_headers = extra_headers or {}
        extra_params = extra_params or {}

        input_params = remove_none(
            {
                "messages": messages,
                "model": model,
                "frequency_penalty": frequency_penalty,
                "function_call": function_call,
                "functions": functions,
                "logit_bias": logit_bias,
                "max_tokens": max_tokens,
                "n": n,
                "presence_penalty": presence_penalty,
                "seed": seed,
                "stop": stop,
                "stream": stream,
                "temperature": temperature,
                "tool_choice": tool_choice,
                "tools": tools,
                "top_p": top_p,
                "user": user,
                "addons": addons,
                "max_prompt_tokens": max_prompt_tokens,
                "custom_fields": custom_fields,
                "logprobs": logprobs,
                "top_logprobs": top_logprobs,
                "extra_body": extra_body,
                "extra_query": {
                    "api-version": (
                        api_version or self.default_api_version or Omit()
                    )
                },
                "extra_headers": {
                    # We use Omit to override openai client auth headers
                    **{"Authorization": Omit(), "api-key": Omit()},
                    **(await self.http_client.auth_headers()),
                    **extra_headers,
                },
            }
        )
        try:
            openai_response = await self.openai_client.chat.completions.create(
                **input_params,
            )
            openai_response = cast(
                Union[
                    OpenaiChatCompletion,
                    OpenaiAsyncStream[OpenaiChatCompletionChunk],
                ],
                openai_response,
            )
        except openai.APIError as err:
            raise convert_openai_error(err)

        if isinstance(openai_response, OpenaiChatCompletion):
            return convert_openai_response(openai_response)
        else:
            return convert_openai_async_stream(openai_response)



from abc import ABC, abstractmethod
from pathlib import PurePosixPath
from typing import Dict, Generic, Optional, TypeVar, Union
from urllib.parse import urljoin

import openai
from httpx import Timeout

import aidial_client.resources as resources
from aidial_client._auth import (
    AsyncAuthValue,
    AuthType,
    AuthValueT,
    SyncAuthValue,
    process_auth,
)
from aidial_client._constants import (
    API_PREFIX,
    DEFAULT_MAX_RETRIES,
    DEFAULT_TIMEOUT,
    OPENAI_PREFIX,
)
from aidial_client._http_client import AsyncHTTPClient, SyncHTTPClient
from aidial_client._internal_types._defaults import NOT_GIVEN, NotGiven
from aidial_client.helpers._url import enforce_trailing_slash
from aidial_client.types.bucket import AppData

_HttpClientT = TypeVar(
    "_HttpClientT", bound=Union[AsyncHTTPClient, SyncHTTPClient]
)


class BaseDialClient(Generic[_HttpClientT, AuthValueT], ABC):
    _auth_type: AuthType
    _auth_value: AuthValueT
    _base_url: str
    _http_client: _HttpClientT
    _auth_headers: Dict[str, str]
    _my_bucket: Optional[str]
    _my_appdata: Union[AppData, None, NotGiven]

    def __init__(
        self,
        *,
        base_url: str,
        api_key: Optional[AuthValueT] = None,
        bearer_token: Optional[AuthValueT] = None,
        max_retries: int = DEFAULT_MAX_RETRIES,
        timeout: Union[float, Timeout, None] = DEFAULT_TIMEOUT,
        api_version: Optional[str] = None,
        http_client: Optional[_HttpClientT] = None,
    ):
        self._auth_type, self._auth_value = process_auth(
            api_key=api_key, bearer_token=bearer_token
        )
        self._max_retries = max_retries
        self._timeout = timeout
        self._base_url = enforce_trailing_slash(base_url)
        self._api_version = api_version
        self._http_client = http_client or self._create_http_client()
        self._my_bucket = None
        self._my_appdata = NOT_GIVEN
        self._init_resources()

    @abstractmethod
    def _init_resources(self) -> None: ...

    @abstractmethod
    def _create_http_client(self) -> _HttpClientT: ...

    def is_dial_url(self, absolute_url: str) -> bool:
        return enforce_trailing_slash(absolute_url).startswith(self._base_url)

    @property
    def api_url(self) -> str:
        return urljoin(self._base_url, API_PREFIX)

    @property
    def base_url(self) -> str:
        return self._base_url

    @property
    def api_version(self) -> Optional[str]:
        return self._api_version


class Dial(BaseDialClient[SyncHTTPClient, SyncAuthValue]):
    def _init_resources(self) -> None:
        openai_client = openai.AzureOpenAI(
            api_key="-",
            api_version="-",
            base_url=urljoin(self._base_url, OPENAI_PREFIX),
            http_client=self._http_client.internal_http_client,
        )
        self.chat = resources.Chat(
            http_client=self._http_client,
            completions=resources.chat.ChatCompletions(
                http_client=self._http_client,
                default_api_version=self.api_version,
                openai_client=openai_client,
            ),
        )
        self.bucket = resources.Bucket(http_client=self._http_client)
        self.metadata = resources.Metadata(http_client=self._http_client)
        self.files = resources.Files(
            http_client=self._http_client,
            metadata=self.metadata,
            dial_api_url=self.api_url,
        )
        self.deployments = resources.Deployments(http_client=self._http_client)
        self.application = resources.Application(http_client=self._http_client)

    def _create_http_client(self) -> SyncHTTPClient:
        return SyncHTTPClient(
            self._base_url,
            self._auth_value,
            self._auth_type,
            self._max_retries,
            self._timeout,
        )

    def _get_my_bucket(self) -> str:
        # Wrapper for convenience of unit tests
        return self.bucket.get_bucket()

    def my_bucket(self) -> str:
        if self._my_bucket is None:
            self._my_bucket = self._get_my_bucket()
        return self._my_bucket

    def my_files_home(self) -> PurePosixPath:
        return "files" / PurePosixPath(self.my_bucket())

    def my_conversations_home(self) -> PurePosixPath:
        return "conversations" / PurePosixPath(self.my_bucket())

    def my_prompts_home(self) -> PurePosixPath:
        return "prompts" / PurePosixPath(self.my_bucket())

    def _get_my_appdata(self) -> Optional[AppData]:
        return self.bucket.get_appdata()

    def my_appdata(self) -> Optional[AppData]:
        if isinstance(self._my_appdata, NotGiven):
            self._my_appdata = self._get_my_appdata()
        return self._my_appdata

    def my_appdata_home(self) -> Optional[PurePosixPath]:
        appdata = self.my_appdata()
        if appdata:
            return PurePosixPath(appdata.raw)
        return None

    def auth_headers(self) -> Dict[str, str]:
        return self._http_client.auth_headers()


class AsyncDial(BaseDialClient[AsyncHTTPClient, AsyncAuthValue]):
    def _init_resources(self) -> None:
        openai_client = openai.AsyncAzureOpenAI(
            # set empty string, we will override
            # it with our client values during request
            api_key="",
            api_version="",
            base_url=urljoin(self._base_url, OPENAI_PREFIX),
            http_client=self._http_client.internal_http_client,
            timeout=self._http_client._timeout,
            max_retries=self._http_client._max_retries,
        )
        self.chat = resources.AsyncChat(
            http_client=self._http_client,
            completions=resources.chat.AsyncChatCompletions(
                http_client=self._http_client,
                default_api_version=self.api_version,
                openai_client=openai_client,
            ),
        )
        self.bucket = resources.AsyncBucket(http_client=self._http_client)
        self.metadata = resources.AsyncMetadata(http_client=self._http_client)
        self.files = resources.AsyncFiles(
            http_client=self._http_client,
            metadata=self.metadata,
            dial_api_url=self.api_url,
        )
        self.deployments = resources.AsyncDeployments(
            http_client=self._http_client
        )
        self.application = resources.AsyncApplication(
            http_client=self._http_client
        )

    def _create_http_client(self) -> AsyncHTTPClient:
        return AsyncHTTPClient(
            self._base_url,
            self._auth_value,
            self._auth_type,
            self._max_retries,
            self._timeout,
        )

    async def _get_my_bucket(self) -> str:
        # Wrapper for convenience of unit tests
        return await self.bucket.get_bucket()

    async def my_bucket(self) -> str:
        if self._my_bucket is None:
            self._my_bucket = await self._get_my_bucket()
        return self._my_bucket

    async def my_files_home(self) -> PurePosixPath:
        return "files" / PurePosixPath(await self.my_bucket())

    async def my_conversations_home(self) -> PurePosixPath:
        return "conversations" / PurePosixPath(await self.my_bucket())

    async def my_prompts_home(self) -> PurePosixPath:
        return "prompts" / PurePosixPath(await self.my_bucket())

    async def _get_my_appdata(self) -> Optional[AppData]:
        return await self.bucket.get_appdata()

    async def my_appdata(self) -> Optional[AppData]:
        if isinstance(self._my_appdata, NotGiven):
            self._my_appdata = await self._get_my_appdata()
        return self._my_appdata

    async def my_appdata_home(self) -> Optional[PurePosixPath]:
        appdata = await self.my_appdata()
        if appdata:
            return PurePosixPath(appdata.raw)
        return None

    async def auth_headers(self) -> Dict[str, str]:
        return await self._http_client.auth_headers()



from pathlib import PurePosixPath
from typing import Literal, Optional, Union
from urllib.parse import urljoin

import httpx

from aidial_client._constants import API_PREFIX
from aidial_client._exception import (
    DialException,
    EtagMismatchError,
    InvalidDialURLError,
    ResourceNotFoundError,
)
from aidial_client._internal_types._generic import NoneType
from aidial_client._internal_types._http_request import (
    FileTypes,
    FinalRequestOptions,
)
from aidial_client._utils._dict import remove_none
from aidial_client.helpers.storage_resource import DialStorageResourceMixin
from aidial_client.resources.base import AsyncResource, Resource
from aidial_client.resources.metadata import AsyncMetadata, Metadata
from aidial_client.types.file import FileDownloadResponse
from aidial_client.types.metadata import FileMetadata


def _files_error_processor(
    http_status_error: httpx.HTTPStatusError,
) -> Optional[DialException]:
    if http_status_error.response.status_code == 412:
        return EtagMismatchError(
            message=http_status_error.response.text,
        )
    elif http_status_error.response.status_code == 404:
        return ResourceNotFoundError(
            message=http_status_error.response.text,
        )
    return None


class Files(Resource, DialStorageResourceMixin):
    metadata: Metadata
    resource_type: str = "files"

    def upload(
        self,
        url: Union[str, PurePosixPath],
        file: FileTypes,
        etag_if_match: Optional[str] = None,
        etag_if_none_match: Optional[Literal["*"]] = None,
    ) -> FileMetadata:
        return self.http_client.request(
            cast_to=FileMetadata,
            options=FinalRequestOptions(
                method="PUT",
                url=urljoin(API_PREFIX, self.get_api_path(str(url))),
                files={"file": file},
                headers=remove_none(
                    {
                        "If-Match": etag_if_match,
                        "If-None-Match": etag_if_none_match,
                    }
                ),
            ),
            on_http_error=_files_error_processor,
        )

    def download(
        self,
        url: Union[str, PurePosixPath],
        etag_if_match: Optional[str] = None,
    ) -> FileDownloadResponse:
        storage_resource = self.get_storage_resource(str(url))
        if storage_resource.filename is None:
            raise InvalidDialURLError("URL points to a directory, not a file")
        response = self.http_client.request(
            cast_to=httpx.Response,
            options=FinalRequestOptions(
                method="GET",
                url=urljoin(API_PREFIX, storage_resource.api_path),
                headers=remove_none(
                    {
                        "If-Match": etag_if_match,
                    }
                ),
            ),
            on_http_error=_files_error_processor,
        )
        return FileDownloadResponse(
            response=response, filename=storage_resource.filename
        )

    def delete(
        self,
        url: Union[str, PurePosixPath],
        etag_if_match: Optional[str] = None,
    ) -> None:
        return self.http_client.request(
            cast_to=NoneType,
            options=FinalRequestOptions(
                method="DELETE",
                url=urljoin(API_PREFIX, self.get_api_path(str(url))),
                headers=remove_none(
                    {
                        "If-Match": etag_if_match,
                    }
                ),
            ),
            on_http_error=_files_error_processor,
        )

    def get_metadata(self, url: Union[str, PurePosixPath]) -> FileMetadata:
        return self.metadata.get(
            resource="files",
            relative_url=self.get_api_path(str(url)),
        )


class AsyncFiles(AsyncResource, DialStorageResourceMixin):
    metadata: AsyncMetadata
    resource_type: str = "files"

    async def upload(
        self,
        url: Union[str, PurePosixPath],
        file: FileTypes,
        etag_if_match: Optional[str] = None,
        etag_if_none_match: Optional[Literal["*"]] = None,
    ) -> FileMetadata:

        return await self.http_client.request(
            cast_to=FileMetadata,
            options=FinalRequestOptions(
                method="PUT",
                url=urljoin(API_PREFIX, self.get_api_path(str(url))),
                files={"file": file},
                headers=remove_none(
                    {
                        "If-Match": etag_if_match,
                        "If-None-Match": etag_if_none_match,
                    }
                ),
            ),
            on_http_error=_files_error_processor,
        )

    async def download(
        self,
        url: Union[str, PurePosixPath],
        etag_if_match: Optional[str] = None,
    ) -> FileDownloadResponse:
        storage_resource = self.get_storage_resource(str(url))
        if storage_resource.filename is None:
            raise InvalidDialURLError("URL points to a directory, not a file")
        response = await self.http_client.request(
            cast_to=httpx.Response,
            options=FinalRequestOptions(
                method="GET",
                url=urljoin(API_PREFIX, storage_resource.api_path),
                headers=remove_none(
                    {
                        "If-Match": etag_if_match,
                    }
                ),
            ),
            on_http_error=_files_error_processor,
        )
        return FileDownloadResponse(
            response=response, filename=storage_resource.filename
        )

    async def delete(
        self,
        url: Union[str, PurePosixPath],
        etag_if_match: Optional[str] = None,
    ) -> None:
        return await self.http_client.request(
            cast_to=NoneType,
            options=FinalRequestOptions(
                method="DELETE",
                url=urljoin(API_PREFIX, self.get_api_path(str(url))),
                headers=remove_none(
                    {
                        "If-Match": etag_if_match,
                    }
                ),
            ),
            on_http_error=_files_error_processor,
        )

    async def get_metadata(
        self, url: Union[str, PurePosixPath]
    ) -> FileMetadata:
        return await self.metadata.get(
            resource="files",
            relative_url=self.get_api_path(str(url)),
        )



from typing import Optional, Union

import httpx

from aidial_client._auth import AsyncAuthValue, SyncAuthValue, process_auth
from aidial_client._client import AsyncDial, Dial
from aidial_client._constants import (
    DEFAULT_CONNECTION_LIMITS,
    DEFAULT_MAX_RETRIES,
    DEFAULT_TIMEOUT,
)
from aidial_client._http_client import AsyncHTTPClient, SyncHTTPClient


class DialClientPool:
    def __init__(
        self,
        *,
        connection_limits: httpx.Limits = DEFAULT_CONNECTION_LIMITS,
        **kwargs,
    ):
        self._internal_http_client = httpx.Client(
            limits=connection_limits, **kwargs
        )

    def create_client(
        self,
        *,
        base_url: str,
        api_key: Optional[SyncAuthValue] = None,
        bearer_token: Optional[SyncAuthValue] = None,
        max_retries: int = DEFAULT_MAX_RETRIES,
        timeout: Union[httpx.Timeout, float] = DEFAULT_TIMEOUT,
    ) -> Dial:
        auth_type, auth_value = process_auth(
            api_key=api_key, bearer_token=bearer_token
        )
        return Dial(
            base_url=base_url,
            api_key=api_key,
            bearer_token=bearer_token,
            http_client=SyncHTTPClient(
                base_url=base_url,
                auth_value=auth_value,
                auth_type=auth_type,
                max_retries=max_retries,
                timeout=timeout,
                internal_http_client=self._internal_http_client,
            ),
        )


class AsyncDialClientPool:
    def __init__(
        self,
        *,
        connection_limits: httpx.Limits = DEFAULT_CONNECTION_LIMITS,
        **kwargs,
    ):
        self._internal_http_client = httpx.AsyncClient(
            limits=connection_limits, **kwargs
        )

    def create_client(
        self,
        *,
        base_url: str,
        api_key: Optional[AsyncAuthValue] = None,
        bearer_token: Optional[AsyncAuthValue] = None,
        max_retries: int = DEFAULT_MAX_RETRIES,
        timeout: Union[httpx.Timeout, float] = DEFAULT_TIMEOUT,
    ) -> AsyncDial:
        auth_type, auth_value = process_auth(
            api_key=api_key, bearer_token=bearer_token
        )
        return AsyncDial(
            base_url=base_url,
            api_key=api_key,
            bearer_token=bearer_token,
            http_client=AsyncHTTPClient(
                base_url=base_url,
                auth_value=auth_value,
                auth_type=auth_type,
                max_retries=max_retries,
                timeout=timeout,
                internal_http_client=self._internal_http_client,
            ),
        )



from http import HTTPStatus
from typing import Mapping, Optional


class DialException(Exception):
    def __init__(
        self,
        message: str,
        status_code: int = 500,
        type: Optional[str] = "runtime_error",
        param: Optional[str] = None,
        code: Optional[str] = None,
        display_message: Optional[str] = None,
    ) -> None:
        self.message = message
        self.status_code = status_code
        self.type = type
        self.param = param
        self.code = code
        self.display_message = display_message

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"message={self.message!r},"
            f"status_code={self.status_code!r},"
            f"type={self.type!r},"
            f"param={self.param!r},"
            f"code={self.code!r},"
            f"display_message={self.display_message!r})"
        )

    def __str__(self) -> str:
        return self.__repr__()

    @classmethod
    def from_error_data(
        cls, status_code: int, error_data: Mapping
    ) -> "DialException":
        message = error_data["message"]
        assert isinstance(message, str)
        return cls(
            message=message,
            status_code=status_code,
            type=error_data.get("type"),
            param=error_data.get("param"),
            code=error_data.get("code"),
            display_message=error_data.get("display_message"),
        )


class InvalidRequestError(DialException):
    def __init__(self, message: str, **kwargs) -> None:
        super().__init__(
            message=message,
            type="invalid_request_error",
            status_code=HTTPStatus.BAD_REQUEST,
            **kwargs,
        )


class InvalidDialURLError(InvalidRequestError):
    pass


class NotDialURLError(InvalidRequestError):
    pass


class ParsingDataError(DialException):
    def __init__(self, message: str, **kwargs) -> None:
        super().__init__(
            message=message,
            status_code=HTTPStatus.UNPROCESSABLE_ENTITY,
            **kwargs,
        )


class EtagMismatchError(DialException):
    def __init__(self, message: str, **kwargs) -> None:
        super().__init__(
            message=message,
            status_code=HTTPStatus.PRECONDITION_FAILED,
            **kwargs,
        )


class ResourceNotFoundError(DialException):
    def __init__(self, message: str, **kwargs) -> None:
        super().__init__(
            message=message,
            status_code=HTTPStatus.NOT_FOUND,
            **kwargs,
        )



from typing import Dict, List, Literal, Optional

from aidial_client._compatibility.pydantic import PYDANTIC_V2
from aidial_client._compatibility.pydantic_v1 import root_validator
from aidial_client._internal_types._model import ExtraAllowModel

if PYDANTIC_V2:
    from pydantic import model_validator


class Attachment(ExtraAllowModel):
    type: Optional[str] = None
    title: Optional[str] = None
    data: Optional[str] = None
    url: Optional[str] = None
    reference_type: Optional[str] = None
    reference_url: Optional[str] = None

    if PYDANTIC_V2:

        @model_validator(mode="before")
        @classmethod
        def validate_data_or_url_v2(cls, values):
            if (
                isinstance(values, dict)
                and "data" not in values
                and "url" not in values
            ):
                raise ValueError("Either data or URL must be provided")
            return values

    else:

        @root_validator(pre=True)
        def validate_data_or_url_v1(cls, values):
            if "data" not in values and "url" not in values:
                raise ValueError("Either data or URL must be provided")
            return values


class CustomContent(ExtraAllowModel):
    attachments: Optional[List[Attachment]] = None
    state: Optional[Dict] = None


class CompletionUsage(ExtraAllowModel):
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int


class FunctionCall(ExtraAllowModel):
    arguments: str
    name: str


class FunctionCallDelta(ExtraAllowModel):
    arguments: Optional[str] = None
    name: Optional[str] = None


class ChatCompletionMessageToolCall(ExtraAllowModel):
    id: str
    function: FunctionCall
    type: Literal["function"]


class ToolCallDelta(ExtraAllowModel):
    index: int
    id: Optional[str] = None
    function: Optional[FunctionCallDelta] = None
    type: Optional[Literal["function"]] = None


class ChatCompletionMessage(ExtraAllowModel):
    role: Literal["assistant"]
    content: Optional[str] = None
    custom_content: Optional[CustomContent] = None
    function_call: Optional[FunctionCall] = None
    tool_calls: Optional[List[ChatCompletionMessageToolCall]] = None


class ChatCompletionMessageDelta(ExtraAllowModel):
    role: Optional[Literal["assistant"]] = None
    content: Optional[str] = None
    custom_content: Optional[CustomContent] = None
    function_call: Optional[FunctionCallDelta] = None
    tool_calls: Optional[List[ToolCallDelta]] = None


class Choice(ExtraAllowModel):
    index: int
    message: ChatCompletionMessage
    finish_reason: Optional[str]


class ChoiceDelta(ExtraAllowModel):
    index: int
    delta: ChatCompletionMessageDelta
    finish_reason: Optional[str] = None


class ChatCompletionResponse(ExtraAllowModel):
    id: str
    object: Literal["chat.completion"]
    choices: List[Choice]
    created: int
    model: Optional[str] = None
    usage: Optional[CompletionUsage] = None


class ChatCompletionChunk(ExtraAllowModel):
    id: str
    object: Literal["chat.completion.chunk"]
    choices: List[ChoiceDelta]
    created: int
    model: Optional[str] = None
    usage: Optional[CompletionUsage] = None



from typing import Any, Dict, List, Literal, Optional, Union

from typing_extensions import TypedDict

from aidial_client.types.chat.addon import Addon
from aidial_client.types.chat.function import (
    FunctionCallSpecParam,
    FunctionParam,
)
from aidial_client.types.chat.request_param import Message, ResponseFormat
from aidial_client.types.chat.tool import ToolCallSpecParam, ToolParam


class ChatCompletionRequestCustomFields(TypedDict, total=False):
    configuration: Optional[Dict[str, Any]]


class ChatCompletionRequest(TypedDict, total=False):
    model: str
    temperature: Optional[float]
    top_p: Optional[float]
    stream: Optional[bool]
    stop: Optional[Union[str, List[str]]]
    max_tokens: Optional[int]
    presence_penalty: Optional[float]
    frequency_penalty: Optional[float]
    logit_bias: Optional[Dict]
    user: Optional[str]
    messages: List[Message]
    data_sources: List[Any]
    n: Optional[int]
    seed: Optional[int]
    logprobs: Optional[bool]
    top_logprobs: Optional[float]
    response_format: Optional[ResponseFormat]
    tools: Optional[List[ToolParam]]
    tool_choice: Optional[Union[Literal["none", "auto"], ToolCallSpecParam]]
    functions: Optional[List[FunctionParam]]
    function_call: Optional[
        Union[Literal["none", "auto"], FunctionCallSpecParam]
    ]
    addons: Optional[Addon]
    max_prompt_tokens: Optional[Union[Literal["infinity"], int]]
    custom_fields: Optional[ChatCompletionRequestCustomFields]



from abc import ABC, abstractmethod
from http import HTTPStatus
from random import uniform
from typing import Dict, Generic, Optional, TypeVar, Union

import httpx

from aidial_client._auth import AuthType, AuthValueT
from aidial_client._constants import INITIAL_RETRY_DELAY, MAX_RETRY_DELAY
from aidial_client._exception import DialException
from aidial_client._internal_types._http_request import FinalRequestOptions
from aidial_client._utils._type_guard import is_mapping
from aidial_client.helpers._url import enforce_trailing_slash

_HttpInternalClientT = TypeVar(
    "_HttpInternalClientT", bound=Union[httpx.Client, httpx.AsyncClient]
)


class BaseHTTPClient(ABC, Generic[_HttpInternalClientT, AuthValueT]):
    _internal_http_client: _HttpInternalClientT
    _auth_value: AuthValueT
    _auth_type: AuthType

    def __init__(
        self,
        base_url: str,
        auth_value: AuthValueT,
        auth_type: AuthType,
        max_retries: int,
        timeout: Union[float, httpx.Timeout, None],
        internal_http_client: Optional[_HttpInternalClientT] = None,
    ):
        self.base_url = httpx.URL(enforce_trailing_slash(base_url))
        self._auth_value = auth_value
        self._auth_type = auth_type
        self._max_retries = max_retries
        self._timeout = timeout
        self._internal_http_client = (
            internal_http_client or self._create_internal_client()
        )

    @abstractmethod
    def _create_internal_client(
        self,
    ) -> _HttpInternalClientT: ...

    def _prepare_url(self, url: str) -> httpx.URL:
        parsed_url = httpx.URL(url)
        if parsed_url.is_relative_url:
            merge_raw_path = (
                self.base_url.raw_path + parsed_url.raw_path.lstrip(b"/")
            )
            return self.base_url.copy_with(raw_path=merge_raw_path.rstrip(b"/"))
        return parsed_url

    def _build_request(
        self,
        options: FinalRequestOptions,
        auth_headers: Dict[str, str],
    ) -> httpx.Request:
        custom_headers = options.headers or {}
        return self._internal_http_client.build_request(
            headers={**auth_headers, **custom_headers},
            method=options.method,
            url=self._prepare_url(options.url),
            params=(
                httpx.QueryParams(options.params) if options.params else None
            ),
            json=options.json_data,
            files=options.files,
            timeout=options.get_timeout(self._timeout),
        )

    def _remaining_retries(
        self, remaining_retries, options: FinalRequestOptions
    ) -> int:
        return (
            remaining_retries
            if remaining_retries is not None
            else options.get_max_retries(self._max_retries)
        )

    def _should_retry(self, response: httpx.Response) -> bool:
        if response.status_code == HTTPStatus.REQUEST_TIMEOUT:
            return True

        if response.status_code == HTTPStatus.CONFLICT:
            return True

        if response.status_code == HTTPStatus.TOO_MANY_REQUESTS:
            return True

        return False

    def _calculate_retry_sleep_seconds(
        self,
        remaining_retries: int,
        options: FinalRequestOptions,
    ) -> float:
        max_retries = options.get_max_retries(self._max_retries)

        nb_retries = max_retries - remaining_retries

        # Apply exponential backoff, but not more than the max.
        sleep_seconds = min(
            INITIAL_RETRY_DELAY * pow(2.0, nb_retries), MAX_RETRY_DELAY
        )
        timeout = sleep_seconds + uniform(-0.5, 0.5)
        return max(0, timeout)

    def _make_dial_error_from_response(
        self,
        response: httpx.Response,
    ) -> DialException:
        if response.is_closed and not response.is_stream_consumed:
            # We can't read the response body as it has been closed
            # before it was read. This can happen if an event hook
            # raises a status error.
            return DialException(
                message="Stream was interrupted",
                status_code=response.status_code,
            )

        try:
            message_data = response.json()
            assert is_mapping(message_data)
            error_data = message_data["error"]
            assert is_mapping(error_data)
            return DialException.from_error_data(
                status_code=response.status_code,
                error_data=error_data,
            )
        except Exception:
            return DialException(
                message=response.text, status_code=response.status_code
            )

    @property
    def internal_http_client(self) -> _HttpInternalClientT:
        return self._internal_http_client



from typing import Dict, List, Literal, Optional

from aidial_client._internal_types._model import ExtraAllowModel


class ScaleSettings(ExtraAllowModel):
    scale_type: Literal["standard"]


class Features(ExtraAllowModel):
    rate: Optional[bool] = None
    tokenize: Optional[bool] = None
    temperature: Optional[bool] = None
    truncate_prompt: Optional[bool] = None
    configuration: Optional[bool] = None
    system_prompt: Optional[bool] = None
    tools: Optional[bool] = None
    seed: Optional[bool] = None
    url_attachments: Optional[bool] = None
    folder_attachments: Optional[bool] = None
    allow_resume: Optional[bool] = None
    parallel_tool_calls: Optional[bool] = None


class DeploymentBase(ExtraAllowModel):
    id: str
    object: str
    owner: Optional[str] = None
    reference: Optional[str] = None
    icon_url: Optional[str] = None
    description: Optional[str] = None
    display_name: Optional[str] = None
    display_version: Optional[str] = None
    status: Optional[str] = None
    created_at: int
    updated_at: Optional[int] = None
    scale_settings: Optional[ScaleSettings] = None
    defaults: Dict = {}
    input_attachment_types: Optional[List[str]] = None
    features: Optional[Features] = None


class Deployment(DeploymentBase):
    object: Literal["deployment", "model"]
    model: str


class DeploymentsResponse(ExtraAllowModel):
    data: List[Deployment]
    object: Literal["list"]



from enum import Enum
from inspect import isawaitable
from typing import (
    Awaitable,
    Callable,
    Dict,
    Optional,
    Tuple,
    TypeVar,
    Union,
    overload,
)

from typing_extensions import assert_never


class AuthType(Enum):
    API_KEY = "API_KEY"
    BEARER = "BEARER"


SyncAuthValue = Union[str, Callable[[], str]]
AsyncAuthValue = Union[SyncAuthValue, Callable[[], Awaitable[str]]]

AuthValueT = TypeVar(
    "AuthValueT",
    bound=Union[SyncAuthValue, AsyncAuthValue],
)


@overload
def get_auth_value(auth_value: SyncAuthValue) -> str: ...


@overload
def get_auth_value(
    auth_value: AsyncAuthValue,
) -> Union[str, Awaitable[str]]: ...


def get_auth_value(
    auth_value: Union[SyncAuthValue, AsyncAuthValue]
) -> Union[str, Awaitable[str]]:
    if isinstance(auth_value, str):
        return auth_value
    elif callable(auth_value):
        return auth_value()
    else:
        assert_never(auth_value)


async def aget_auth_value(auth_value: AsyncAuthValue) -> str:
    processed_auth_value = get_auth_value(auth_value)
    if isawaitable(processed_auth_value):
        return await processed_auth_value
    return processed_auth_value


def _get_auth_headers(auth_type: AuthType, auth_value: str) -> Dict[str, str]:
    if auth_type == AuthType.API_KEY:
        return {"api-key": auth_value}
    elif auth_type == AuthType.BEARER:
        return {"Authorization": f"Bearer {auth_value}"}
    else:
        assert_never(auth_type)


def get_auth_headers(
    *,
    auth_value: SyncAuthValue,
    auth_type: AuthType,
) -> Dict[str, str]:
    processed_auth_value = get_auth_value(auth_value)
    return _get_auth_headers(auth_type, processed_auth_value)


async def aget_auth_headers(
    auth_value: AsyncAuthValue,
    auth_type: AuthType,
) -> Dict[str, str]:
    processed_auth_value = await aget_auth_value(auth_value)
    return _get_auth_headers(auth_type, processed_auth_value)


def process_auth(
    *,
    api_key: Optional[AuthValueT] = None,
    bearer_token: Optional[AuthValueT] = None,
) -> Tuple[AuthType, AuthValueT]:
    if api_key and bearer_token:
        raise ValueError(
            "Either api_key or bearer_token must be provided, but not both"
        )
    elif api_key:
        return AuthType.API_KEY, api_key
    elif bearer_token:
        return AuthType.BEARER, bearer_token
    else:
        raise ValueError("Either api_key or bearer_token must be provided")



from typing import Literal

from typing_extensions import Required, TypedDict

from aidial_client.types.chat.function import (
    FunctionCallParam,
    FunctionCallSpecParam,
    FunctionParam,
)


class ToolParam(TypedDict):
    type: Literal["function"]
    function: FunctionParam


class ToolCallParam(TypedDict):
    id: Required[str]
    type: Required[Literal["function"]]
    function: FunctionCallParam


class ToolCallSpecParam(TypedDict, total=False):
    type: Required[Literal["function"]]
    function: FunctionCallSpecParam



from typing import Literal


class NotGiven:
    """
    A sentinel singleton class used to distinguish omitted keyword arguments
    from those passed in with the value None
    (which may have different behavior).
    """

    def __bool__(self) -> Literal[False]:
        return False


NOT_GIVEN = NotGiven()



from urllib.parse import urljoin

import httpx

DEFAULT_MAX_RETRIES = 2
DEFAULT_TIMEOUT = httpx.Timeout(timeout=600.0, connect=5.0)
DEFAULT_CONNECTION_LIMITS = httpx.Limits(
    max_connections=1000, max_keepalive_connections=100
)
INITIAL_RETRY_DELAY = 0.5
MAX_RETRY_DELAY = 8.0
API_PREFIX = "v1/"
METADATA_PREFIX = urljoin(API_PREFIX, "metadata/")
FILES_PREFIX = urljoin(API_PREFIX, "files/")


OPENAI_PREFIX = "openai/"
APPLICATION_PREFIX = urljoin(OPENAI_PREFIX, "applications/")



from typing import Dict, Optional

from typing_extensions import Required, TypedDict


class FunctionParam(TypedDict, total=False):
    name: Required[str]
    description: Optional[str]
    parameters: Optional[Dict]


class FunctionCallParam(TypedDict):
    name: Required[str]
    arguments: Required[str]


class FunctionCallSpecParam(TypedDict):
    name: Required[str]



from aidial_client.resources.base import AsyncResource, Resource
from aidial_client.resources.chat.completions import (
    AsyncChatCompletions,
    ChatCompletions,
)


class Chat(Resource):
    completions: ChatCompletions


class AsyncChat(AsyncResource):
    completions: AsyncChatCompletions



from typing import Dict, List, Literal, Optional

from aidial_client._internal_types._model import ExtraAllowModel
from aidial_client.types.deployment import DeploymentBase


class Application(DeploymentBase):
    object: Literal["application"]
    application: str
    application_type_schema_id: Optional[str] = None
    application_properties: Optional[Dict] = None
    invalid: Optional[bool] = None


class ApplicationsResponse(ExtraAllowModel):
    data: List[Application]
    object: Literal["list"]



from typing import Type, cast

import httpx

from aidial_client._exception import ParsingDataError
from aidial_client._internal_types._generic import NoneType, ResponseT
from aidial_client._internal_types._model import (
    ExtraAllowModel,
    ExtraForbidModel,
)


def process_block_response(
    cast_to: Type[ResponseT], response: httpx.Response
) -> ResponseT:
    if cast_to == httpx.Response:
        return cast(ResponseT, response)
    elif cast_to == bytes:
        return cast(ResponseT, response.content)
    elif cast_to == str:
        return cast(ResponseT, response.text)
    elif cast_to == NoneType:
        return cast(ResponseT, None)
    elif issubclass(cast_to, (ExtraForbidModel, ExtraAllowModel)):
        try:
            data = response.json()
            return cast_to(**data)
        except Exception as e:
            raise ParsingDataError(
                message=f"Error during parsing of response data: {str(e)}"
            )
    else:
        raise NotImplementedError("This cast_to type is not supported.")



from typing import List
from urllib.parse import urljoin

from aidial_client._constants import APPLICATION_PREFIX
from aidial_client._internal_types._http_request import FinalRequestOptions
from aidial_client.resources.base import AsyncResource, Resource
from aidial_client.types.application import Application as ApplicationType
from aidial_client.types.application import ApplicationsResponse


class Application(Resource):
    def get(self, app_id: str) -> ApplicationType:
        return self.http_client.request(
            cast_to=ApplicationType,
            options=FinalRequestOptions(
                method="GET", url=urljoin(APPLICATION_PREFIX, app_id)
            ),
        )

    def _list_raw(self) -> ApplicationsResponse:
        return self.http_client.request(
            cast_to=ApplicationsResponse,
            options=FinalRequestOptions(method="GET", url=APPLICATION_PREFIX),
        )

    def list(self) -> List[ApplicationType]:
        return self._list_raw().data


class AsyncApplication(AsyncResource):
    async def get(self, app_id: str) -> ApplicationType:
        return await self.http_client.request(
            cast_to=ApplicationType,
            options=FinalRequestOptions(
                method="GET",
                url=urljoin(APPLICATION_PREFIX, app_id),
            ),
        )

    async def _list_raw(self) -> ApplicationsResponse:
        return await self.http_client.request(
            cast_to=ApplicationsResponse,
            options=FinalRequestOptions(method="GET", url=APPLICATION_PREFIX),
        )

    async def list(self) -> List[ApplicationType]:
        return (await self._list_raw()).data



from aidial_client._compatibility.pydantic_v1 import BaseModel
from aidial_client._http_client import AsyncHTTPClient, SyncHTTPClient


class Resource(BaseModel):
    class Config:
        arbitrary_types_allowed = True

    http_client: SyncHTTPClient


class AsyncResource(BaseModel):
    class Config:
        arbitrary_types_allowed = True

    http_client: AsyncHTTPClient



from typing import Optional
from urllib.parse import urljoin

from aidial_client._constants import API_PREFIX
from aidial_client._internal_types._http_request import FinalRequestOptions
from aidial_client.resources.base import AsyncResource, Resource
from aidial_client.types.bucket import AppData, BucketResponse


class Bucket(Resource):
    def get_raw(self) -> BucketResponse:
        return self.http_client.request(
            cast_to=BucketResponse,
            options=FinalRequestOptions(
                method="GET", url=urljoin(API_PREFIX, "bucket")
            ),
        )

    def get_bucket(self) -> str:
        response = self.get_raw()
        return response.bucket

    def get_appdata(self) -> Optional[AppData]:
        response = self.get_raw()
        if not response.appdata:
            return None
        return AppData.parse(response.appdata)


class AsyncBucket(AsyncResource):
    async def get_raw(self) -> BucketResponse:
        return await self.http_client.request(
            cast_to=BucketResponse,
            options=FinalRequestOptions(
                method="GET", url=urljoin(API_PREFIX, "bucket")
            ),
        )

    async def get_bucket(self) -> str:
        response = await self.get_raw()
        return response.bucket

    async def get_appdata(self) -> Optional[AppData]:
        response = await self.get_raw()
        if not response.appdata:
            return None
        return AppData.parse(response.appdata)



from typing import Literal, Type, Union, overload
from urllib.parse import urljoin

from typing_extensions import assert_never

from aidial_client._constants import METADATA_PREFIX
from aidial_client._internal_types._http_request import FinalRequestOptions
from aidial_client.helpers.storage_resource import StorageResourceType
from aidial_client.resources.base import AsyncResource, Resource
from aidial_client.types.metadata import (
    ConversationMetadata,
    FileMetadata,
    PromptMetadata,
)


def _get_cast_to(
    resource: StorageResourceType,
) -> Union[
    Type[FileMetadata], Type[ConversationMetadata], Type[PromptMetadata]
]:
    if resource == "files":
        return FileMetadata
    elif resource == "conversations":
        return ConversationMetadata
    elif resource == "prompts":
        return PromptMetadata
    else:
        assert_never(resource)


class Metadata(Resource):
    @overload
    def get(
        self, resource: Literal["files"], relative_url: str
    ) -> FileMetadata: ...

    @overload
    def get(
        self, resource: Literal["conversations"], relative_url: str
    ) -> ConversationMetadata: ...

    @overload
    def get(
        self, resource: Literal["prompts"], relative_url: str
    ) -> PromptMetadata: ...

    def get(
        self,
        resource: StorageResourceType,
        relative_url: str,
    ) -> Union[FileMetadata, ConversationMetadata, PromptMetadata]:
        return self.http_client.request(
            cast_to=_get_cast_to(resource),
            options=FinalRequestOptions(
                method="GET",
                url=urljoin(METADATA_PREFIX, relative_url),
            ),
        )


class AsyncMetadata(AsyncResource):
    @overload
    async def get(
        self, resource: Literal["files"], relative_url: str
    ) -> FileMetadata: ...

    @overload
    async def get(
        self, resource: Literal["conversations"], relative_url: str
    ) -> ConversationMetadata: ...

    @overload
    async def get(
        self, resource: Literal["prompts"], relative_url: str
    ) -> PromptMetadata: ...

    async def get(
        self,
        resource: StorageResourceType,
        relative_url: str,
    ) -> Union[FileMetadata, ConversationMetadata, PromptMetadata]:
        return await self.http_client.request(
            cast_to=_get_cast_to(resource),
            options=FinalRequestOptions(
                method="GET",
                url=urljoin(METADATA_PREFIX, relative_url),
            ),
        )



from aidial_client._auth import AsyncAuthValue, AuthType, SyncAuthValue
from aidial_client._client import AsyncDial, Dial
from aidial_client._client_pool import AsyncDialClientPool, DialClientPool
from aidial_client._exception import (
    DialException,
    EtagMismatchError,
    InvalidDialURLError,
    InvalidRequestError,
    ParsingDataError,
    ResourceNotFoundError,
)

__all__ = [
    "Dial",
    "AsyncDial",
    "DialClientPool",
    "AsyncDialClientPool",
    "AuthType",
    "SyncAuthValue",
    "AsyncAuthValue",
    # Exceptions
    "DialException",
    "InvalidDialURLError",
    "InvalidRequestError",
    "ParsingDataError",
    "EtagMismatchError",
    "ResourceNotFoundError",
]



from enum import Enum
from typing import Any, Dict, List, Literal, Mapping, Optional, Union

from aidial_client._compatibility.pydantic_v1 import (
    ConstrainedFloat,
    ConstrainedInt,
    ConstrainedList,
    PositiveInt,
    StrictBool,
    StrictInt,
    StrictStr,
)
from aidial_client._internal_types._model import ExtraForbidModel


class FinishReason(str, Enum):
    STOP = "stop"
    LENGTH = "length"
    FUNCTION_CALL = "function_call"
    TOOL_CALLS = "tool_calls"
    CONTENT_FILTER = "content_filter"


class Status(str, Enum):
    COMPLETED = "completed"
    FAILED = "failed"


class Attachment(ExtraForbidModel):
    type: Optional[StrictStr] = "text/markdown"
    title: Optional[StrictStr] = None
    data: Optional[StrictStr] = None
    url: Optional[StrictStr] = None
    reference_type: Optional[StrictStr] = None
    reference_url: Optional[StrictStr] = None


class Stage(ExtraForbidModel):
    name: StrictStr
    status: Status
    content: Optional[StrictStr] = None
    attachments: Optional[List[Attachment]] = None


class CustomContent(ExtraForbidModel):
    stages: Optional[List[Stage]] = None
    attachments: Optional[List[Attachment]] = None
    state: Optional[Any] = None


class FunctionCall(ExtraForbidModel):
    name: str
    arguments: str


class ToolCall(ExtraForbidModel):
    # OpenAI API doesn't strictly specify existence of the index field
    index: Optional[int]
    id: StrictStr
    type: Literal["function"]
    function: FunctionCall


class Role(str, Enum):
    SYSTEM = "system"
    USER = "user"
    ASSISTANT = "assistant"
    FUNCTION = "function"
    TOOL = "tool"


class Message(ExtraForbidModel):
    role: Role
    content: Optional[StrictStr] = None
    custom_content: Optional[CustomContent] = None
    name: Optional[StrictStr] = None
    tool_calls: Optional[List[ToolCall]] = None
    tool_call_id: Optional[StrictStr] = None
    function_call: Optional[FunctionCall] = None


class Addon(ExtraForbidModel):
    name: Optional[StrictStr] = None
    url: Optional[StrictStr] = None


class Function(ExtraForbidModel):
    name: StrictStr
    description: Optional[StrictStr] = None
    parameters: Optional[Dict] = None


class Temperature(ConstrainedFloat):
    ge = 0
    le = 2


class TopP(ConstrainedFloat):
    ge = 0
    le = 1


class N(ConstrainedInt):
    ge = 1
    le = 128


class Stop(ConstrainedList):
    max_items: int = 4
    __args__ = tuple([StrictStr])


class Penalty(ConstrainedFloat):
    ge = -2
    le = 2


class Tool(ExtraForbidModel):
    type: Literal["function"]
    function: Function


class FunctionChoice(ExtraForbidModel):
    name: StrictStr


class ToolChoice(ExtraForbidModel):
    type: Literal["function"]
    function: FunctionChoice


class ResponseFormat(ExtraForbidModel):
    type: Literal["text", "json_object"]


class AzureChatCompletionRequest(ExtraForbidModel):
    model: Optional[StrictStr] = None
    messages: List[Message]
    functions: Optional[List[Function]] = None
    function_call: Optional[Union[Literal["auto", "none"], FunctionChoice]] = (
        None
    )
    tools: Optional[List[Tool]] = None
    tool_choice: Optional[Union[Literal["auto", "none"], ToolChoice]] = None
    stream: bool = False
    temperature: Optional[Temperature] = None
    top_p: Optional[TopP] = None
    n: Optional[N] = None
    stop: Optional[Union[StrictStr, Stop]] = None
    max_tokens: Optional[PositiveInt] = None
    presence_penalty: Optional[Penalty] = None
    frequency_penalty: Optional[Penalty] = None
    logit_bias: Optional[Mapping[int, float]] = None
    user: Optional[StrictStr] = None
    seed: Optional[StrictInt] = None
    logprobs: Optional[StrictBool] = None
    top_logprobs: Optional[StrictInt] = None
    response_format: Optional[ResponseFormat] = None


class ChatCompletionRequestCustomFields(ExtraForbidModel):
    configuration: Optional[Dict[str, Any]] = None


class ChatCompletionRequest(AzureChatCompletionRequest):
    addons: Optional[List[Addon]] = None
    max_prompt_tokens: Optional[PositiveInt] = None
    custom_fields: Optional[ChatCompletionRequestCustomFields] = None



from typing import Dict, List, Literal, Optional, Union

from typing_extensions import Required, TypedDict

from aidial_client.types.chat.function import FunctionCallParam
from aidial_client.types.chat.tool import ToolCallParam


class ResponseFormat(TypedDict, total=False):
    type: Literal["json_object", "text"]


class AttachmentParam(TypedDict, total=False):
    type: str
    title: str
    data: str
    url: str
    reference_type: str
    reference_url: str


class CustomContentParam(TypedDict, total=False):
    attachments: Optional[List[AttachmentParam]]
    state: Optional[Dict]


class SystemMessageParam(TypedDict, total=False):
    role: Required[Literal["system"]]
    content: Required[str]
    custom_content: Optional[CustomContentParam]
    name: Optional[str]


class UserMessageParam(TypedDict, total=False):
    role: Required[Literal["user"]]
    content: Required[str]
    custom_content: Optional[CustomContentParam]
    name: Optional[str]


class AssistantMessageParam(TypedDict, total=False):
    role: Required[Literal["assistant"]]
    content: Optional[str]
    custom_content: Optional[CustomContentParam]
    function_call: Optional[FunctionCallParam]
    tool_calls: List[ToolCallParam]
    name: Optional[str]


class ToolMessageParam(TypedDict, total=False):
    role: Required[Literal["tool"]]
    content: Required[str]
    tool_call_id: Required[str]


class FunctionMessageParam(TypedDict, total=False):
    role: Required[Literal["function"]]
    content: Required[str]
    """Name of function call"""
    name: Required[str]


Message = Union[
    SystemMessageParam,
    UserMessageParam,
    AssistantMessageParam,
    ToolMessageParam,
    FunctionMessageParam,
]



import pydantic

from aidial_client._compatibility.openai import BaseModel as OpenAIBaseModel
from aidial_client._compatibility.pydantic import PYDANTIC_V2
from aidial_client._compatibility.pydantic_v1 import BaseModel, Extra


class ExtraAllowModel(OpenAIBaseModel):
    if PYDANTIC_V2:
        model_config = {"extra": "allow"}
    else:

        class Config(pydantic.BaseConfig):  # pyright: ignore[reportDeprecated]
            extra = Extra.allow


class ExtraForbidModel(BaseModel):
    class Config:
        extra = Extra.forbid  # type: ignore



"""
Since we need some protected imports from openai, wrap it with this module,
for easier handling of cases, when such member will migrate to another modules
"""

from openai._models import BaseModel  # noqa: F401
from openai._types import Omit  # noqa: F401



from typing import Union

from typing_extensions import TypedDict


class ExternalAddon(TypedDict):
    # The URL for accessing the OpenAI Plugin Schema.
    # The system object used for converting the Addon name to the Addon link.
    url: str


class SystemAddon(TypedDict):
    # The name of the system Addon.
    name: str


Addon = Union[ExternalAddon, SystemAddon]



from pathlib import Path
from typing import Union

import aiofiles
import httpx


class FileDownloadResponse:

    def __init__(self, response: httpx.Response, filename: str):
        self._response = response
        self._filename = filename

    def write_to(self, file: Union[str, Path]) -> None:
        """
        Write the content to a file
        """
        with open(file, "wb") as f:
            for chunk in self._response.iter_bytes():
                f.write(chunk)

    async def awrite_to(self, file: Union[str, Path]) -> None:
        """
        Async write content to a file
        """

        async with aiofiles.open(file, "wb") as f:
            async for chunk in self._response.aiter_bytes():
                await f.write(chunk)

    def __aiter__(self):
        return self._response.aiter_bytes()

    def __iter__(self):
        return self._response.iter_bytes()

    def get_content(self) -> bytes:
        return self._response.read()

    async def aget_content(self) -> bytes:
        return await self._response.aread()

    @property
    def filename(self) -> str:
        return self._filename



from typing import AsyncIterator, Iterator

import openai
from openai.types.chat import ChatCompletion as OpenAIChatCompletion
from openai.types.chat import ChatCompletionChunk as OpenAIChatCompletionChunk

from aidial_client._exception import DialException
from aidial_client.types.chat import ChatCompletionChunk, ChatCompletionResponse


def convert_openai_error(error: openai.APIError) -> DialException:
    status_code = (
        error.status_code if isinstance(error, openai.APIStatusError) else 500
    )
    display_message = None
    if (
        hasattr(error, "body")
        and error.body is not None
        and isinstance(error.body, dict)
    ):
        display_message = error.body.get("display_message", None)
    return DialException(
        message=error.message,
        status_code=status_code,
        type=error.type,
        param=error.param,
        code=error.code,
        display_message=display_message,
    )


def convert_openai_response(
    openai_response: OpenAIChatCompletion,
) -> ChatCompletionResponse:
    return ChatCompletionResponse(**openai_response.model_dump())


def convert_openai_stream(
    openai_response: Iterator[OpenAIChatCompletionChunk],
) -> Iterator[ChatCompletionChunk]:
    try:
        for chunk in openai_response:
            yield ChatCompletionChunk(**chunk.model_dump())
    except openai.APIError as e:
        raise convert_openai_error(e) from e


async def convert_openai_async_stream(
    openai_response: AsyncIterator[OpenAIChatCompletionChunk],
) -> AsyncIterator[ChatCompletionChunk]:
    try:
        async for chunk in openai_response:
            yield ChatCompletionChunk(**chunk.model_dump())
    except openai.APIError as e:
        raise convert_openai_error(e) from e



import asyncio
from http import HTTPStatus
from typing import Callable, Dict, Optional, Type

import httpx

from aidial_client._auth import AsyncAuthValue, aget_auth_headers
from aidial_client._exception import DialException
from aidial_client._http_client._base import BaseHTTPClient
from aidial_client._internal_types._generic import ResponseT
from aidial_client._internal_types._http_request import FinalRequestOptions
from aidial_client._log import logger
from aidial_client._utils._response_processing import process_block_response


class AsyncHTTPClient(BaseHTTPClient[httpx.AsyncClient, AsyncAuthValue]):
    def _create_internal_client(self) -> httpx.AsyncClient:
        return httpx.AsyncClient(
            timeout=self._timeout,
        )

    async def auth_headers(self) -> Dict[str, str]:
        return await aget_auth_headers(
            auth_value=self._auth_value, auth_type=self._auth_type
        )

    async def _retry_request(
        self,
        *,
        options: FinalRequestOptions,
        cast_to: Type[ResponseT],
        remaining_retries: int,
    ) -> ResponseT:
        remaining = remaining_retries - 1
        logger.debug(f"Retries left: {remaining}")

        sleep_time = self._calculate_retry_sleep_seconds(remaining, options)
        logger.info(f"Making retry to {options.url} in {sleep_time} seconds")
        await asyncio.sleep(sleep_time)

        return await self.request(
            options=options, cast_to=cast_to, remaining_retries=remaining
        )

    async def request(
        self,
        *,
        options: FinalRequestOptions,
        cast_to: Type[ResponseT],
        remaining_retries: Optional[int] = None,
        on_http_error: Optional[
            Callable[[httpx.HTTPStatusError], Optional[DialException]]
        ] = None,
    ) -> ResponseT:
        retries = self._remaining_retries(remaining_retries, options)
        auth_headers = await self.auth_headers()

        request = self._build_request(options, auth_headers)
        try:
            response = await self._internal_http_client.send(request)
        except httpx.TimeoutException as err:
            logger.debug("Request failed by timeout")

            if retries > 0:
                return await self._retry_request(
                    options=options,
                    cast_to=cast_to,
                    remaining_retries=retries,
                )

            raise DialException(
                message="Request timed out",
                status_code=HTTPStatus.REQUEST_TIMEOUT,
            ) from err
        except Exception as err:
            logger.debug("Unknown exception")
            if retries > 0:
                return await self._retry_request(
                    options=options,
                    cast_to=cast_to,
                    remaining_retries=retries,
                )
            raise DialException(message="Unknown error during request") from err

        logger.debug(f"HTTP Response received with {response.status_code}")

        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as err:
            logger.debug(
                f"Encountered error HTTP status: {err.response.status_code}"
                f"Content: {err.response.text}"
            )

            if retries > 0 and self._should_retry(err.response):
                err.response.close()
                return await self._retry_request(
                    options=options,
                    cast_to=cast_to,
                    remaining_retries=retries,
                )
            # Try to get custom error from response status_code/code/message
            custom_error = on_http_error(err) if on_http_error else None
            # or fallback to default processing
            raised_error = custom_error or self._make_dial_error_from_response(
                err.response
            )
            raise raised_error from err

        return process_block_response(cast_to=cast_to, response=response)



import re
from typing import Optional

from aidial_client._internal_types._model import ExtraAllowModel


class AppData(ExtraAllowModel):
    raw: str
    user_bucket: str
    app_name: str

    @classmethod
    def parse(cls, appdata: str) -> "AppData":
        match = re.match(r"^(.+)/appdata/(.+)$", appdata)
        if not match:
            raise ValueError("Invalid appdata format")

        user_bucket, app_name = match.groups()
        return cls(raw=appdata, user_bucket=user_bucket, app_name=app_name)


class BucketResponse(ExtraAllowModel):
    bucket: str
    appdata: Optional[str] = None



from typing import Dict, Mapping, Optional

from aidial_client._auth import AuthType, get_auth_headers
from aidial_client._compatibility.pydantic_v1 import (
    SecretStr,
    StrictStr,
    root_validator,
)
from aidial_client._internal_types._model import ExtraForbidModel
from aidial_client.types.chat.legacy.chat_completion import (
    ChatCompletionRequest,
)


class RequestParams(ExtraForbidModel):
    api_key_secret: SecretStr
    jwt_secret: Optional[SecretStr] = None

    deployment_id: StrictStr
    api_version: Optional[StrictStr] = None
    headers: Mapping[StrictStr, StrictStr]

    @root_validator(pre=True)
    def create_secrets(cls, values: dict):
        if "api_key" in values:
            if "api_key_secret" not in values:
                values["api_key_secret"] = SecretStr(values.pop("api_key"))
            else:
                raise ValueError(
                    "api_key and api_key_secret cannot be both provided"
                )

        if "jwt" in values:
            if "jwt_secret" not in values:
                values["jwt_secret"] = SecretStr(values.pop("jwt"))
            else:
                raise ValueError("jwt and jwt_secret cannot be both provided")

        return values

    @property
    def api_key(self) -> str:
        return self.api_key_secret.get_secret_value()

    @property
    def jwt(self) -> Optional[str]:
        return self.jwt_secret.get_secret_value() if self.jwt_secret else None

    @property
    def auth_headers(self) -> Dict[str, str]:
        if self.jwt_secret is not None:
            return get_auth_headers(
                auth_type=AuthType.BEARER,
                auth_value=self.jwt_secret.get_secret_value(),
            )
        else:
            return get_auth_headers(
                auth_type=AuthType.API_KEY,
                auth_value=self.api_key_secret.get_secret_value(),
            )


class ApplicationChatCompletionRequest(ChatCompletionRequest, RequestParams):
    pass



from typing import TypeVar, Union

import httpx

from aidial_client._internal_types._model import (
    ExtraAllowModel,
    ExtraForbidModel,
)
from aidial_client.types.file import FileDownloadResponse

ResponseT = TypeVar(
    "ResponseT",
    bound=Union[
        ExtraAllowModel,
        ExtraForbidModel,
        bytes,
        str,
        httpx.Response,
        FileDownloadResponse,
        None,
    ],
)
NoneType = type(None)



"""
Just copy of alias generators from pydantic V2:
https://github.com/pydantic/pydantic/blob/c772b43edb952c5fe54bb28da5124b10d5470caf/pydantic/alias_generators.py

So we can use library  with pydantic < 2.0 version
"""

import re


def to_pascal(snake: str) -> str:
    """Convert a snake_case string to PascalCase.

    Args:
        snake: The string to convert.

    Returns:
        The PascalCase string.
    """
    camel = snake.title()
    return re.sub("([0-9A-Za-z])_(?=[0-9A-Z])", lambda m: m.group(1), camel)


def to_camel(snake: str) -> str:
    """Convert a snake_case string to camelCase.

    Args:
        snake: The string to convert.

    Returns:
        The converted camelCase string.
    """
    # If the string is already in camelCase
    # and does not contain a digit followed
    # by a lowercase letter, return it as it is
    if re.match("^[a-z]+[A-Za-z0-9]*$", snake) and not re.search(
        r"\d[a-z]", snake
    ):
        return snake

    camel = to_pascal(snake)
    return re.sub("(^_*[A-Z])", lambda m: m.group(1).lower(), camel)



from .addon import Addon, ExternalAddon, SystemAddon
from .function import FunctionCallSpecParam, FunctionParam
from .request import ChatCompletionRequest
from .request_param import (
    FunctionMessageParam,
    Message,
    SystemMessageParam,
    ToolMessageParam,
    UserMessageParam,
)
from .response import ChatCompletionChunk, ChatCompletionResponse
from .tool import ToolCallSpecParam, ToolParam

__all__ = [
    "ChatCompletionRequest",
    "Addon",
    "SystemAddon",
    "ExternalAddon",
    "FunctionParam",
    "FunctionCallSpecParam",
    "ToolParam",
    "ToolCallSpecParam",
    "Message",
    "ToolMessageParam",
    "UserMessageParam",
    "SystemMessageParam",
    "FunctionMessageParam",
    "ChatCompletionResponse",
    "ChatCompletionChunk",
]



def enforce_trailing_slash(url: str) -> str:
    if url.endswith("/"):
        return url
    return url + "/"


def remove_prefix(s: str, prefix: str) -> str:
    if s.startswith(prefix):
        s = s.lstrip(prefix)
    return s


def remove_leading_slash(url: str) -> str:
    return remove_prefix(url, "/")



from typing import List, Literal, Optional

from aidial_client._compatibility.pydantic import PYDANTIC_V2
from aidial_client._internal_types._model import ExtraAllowModel
from aidial_client._utils._alias import to_camel


class BaseMetadata(ExtraAllowModel):
    if PYDANTIC_V2:
        model_config = {
            "alias_generator": to_camel,
            "populate_by_name": True,
        }
    else:

        class Config:
            alias_generator = to_camel
            allow_population_by_field_name = True

    name: str
    parent_path: Optional[str] = None
    bucket: str
    url: str
    node_type: Literal["FOLDER", "ITEM"]
    resource_type: Literal["FILE", "CONVERSATION", "PROMPT"]


class FileItem(BaseMetadata):
    node_type: Literal["FOLDER", "ITEM"]
    resource_type: Literal["FILE"]
    content_length: Optional[int] = None
    content_type: Optional[str] = None


class FileMetadata(BaseMetadata):
    node_type: Literal["FOLDER", "ITEM"]
    resource_type: Literal["FILE"]
    content_length: Optional[int] = None
    content_type: Optional[str] = None
    items: Optional[List[FileItem]] = None
    etag: Optional[str] = None


class ConversationItem(BaseMetadata):
    updated_at: int
    resource_type: Literal["CONVERSATION"]


class ConversationMetadata(BaseMetadata):
    content_length: Optional[int] = None
    next_token: Optional[str] = None
    items: Optional[List[ConversationItem]]
    resource_type: Literal["CONVERSATION"]


class PromptItem(BaseMetadata):
    updated_at: int
    resource_type: Literal["PROMPT"]


class PromptMetadata(BaseMetadata):
    content_length: Optional[int] = None
    next_token: Optional[str] = None
    items: Optional[List[PromptItem]]
    resource_type: Literal["PROMPT"]
